/*****
 *
 * TBR20A.PRG
 *
 * Clipper tbr20 /N
 * RTLINK FILE tbr20a
 *
 This CA-Demo-Programm is from CLIPPER 5.2E
 I have modified it to show how WILDSRCH works.

 Modifications by:    Rainer Kppl-Schfer
                      RKS-Software
                      Goldbergstrasse 9
                      D-71065 Sindelfingen


 The Changes are:

 - Calling BroWhile-Function with Dummy-Condition (.T.) - to browse all
   records

 - New functions for WildSrch : wGoTopDB      (original: GoTopDB)
                                wGoBottomDB   (original: GoBottomDB)
                                wSkipDB       (original: SkipDB)
                                The "w"-Functions were called instead
                                of the original functions, when WILDSRCH
                                is active (lWildSrch:=.t.)

 - Hotkey to call the WILDSRCH-Function is <Alt+W>.
   Press <Alt+W> to enter the SearchArgument
   if WILDSRCH is active press <AltW> again to return to "normal" browse.


 The included TEST.DBF holds 400 records. The file will be indexed on <FLD3>.
 To get more records for your test do:

                     use test
                     copy to xx
                     append from  xx ..   800 records
                     append from  xx ..  1200   "
                     append from  xx ..  1600   "
                     append from  xx ..  2000   "   ... and so on..
 */

#include "inkey.ch"
#include "setcurs.ch"

#define     COLSEP     CHR(32)  + CHR(179) + CHR(32)
#define     HEADSEP    CHR(205) + CHR(209) + CHR(205)
#define     FOOTSEP    CHR(205) + CHR(207) + CHR(205)

// WildSrch
EXTERNAL ASCAN

FUNCTION Main()
   LOCAL cScreen
   // Serves just for test purposes
   cScreen := savescreen()
   SETCURSOR(SC_NONE)
   SCROLL()

   // Open File
   if !file("test.ntx")
      USE test NEW
      @ 10,10 say "Indexing, please wait.."
      index on upper(fld3) to TEST
      USE
   endif
   // Open file in shared Mode
   USE test NEW SHARED
   DBSETINDEX("test")

   BroWhile( ,,,,,, ".T." )     // Dummy-Condition, browses all records
   SCROLL()
   restscreen(,,,,cScreen)
   RETURN (NIL)

/*****
 *
 * BroWhile()
 *
 * nTop, nLeft, nBottom, nRight - window coordinates
 *                                defaults to min and max
 *                                of the screen
 * aColFld - array of columns for table
 *           (defaults to all fields)
 * cUdf - UDF label - passed as string
 * cWhileCond - while condition - passed as string
 * cColTit - array of titles for each column
 *
 * This generic function is based on the TBrowse Class.
 * Made to act like DBEDIT() and include its significant
 * parameters only. Most importantly, it incorporates a
 * WHILE condition.
 *
 * BroWhile( , , , , , , cFiltCond )        // do it!
 * BroWhile( , , , , , "Udf" , cFiltCond )  // do it
 *                                             with Udf!
 *
 *
 */

FUNCTION BroWhile( nTop, nLeft, nRight, nBottom,;
                   aColFld, cUdf, cWhileCond, aColTit )

   LOCAL bWhileCond, bUdf
   LOCAL cColr := SETCOLOR()
   LOCAL lColFld, lColTit, lUdf
   LOCAL nCOunt, nRow, nCol, nKey := 0
   LOCAL nRetUdf := 1
   LOCAL nNumOfFlds, nCursor := SETCURSOR(SC_NONE)
   LOCAL nMode := 4
   LOCAL o, column

   // WildSrch
   LOCAL wScreen
   LOCAL cSrchArg:=space(20)
   LOCAL lWildSrch := .f.
   PRIVATE aResult:={}

   // WildSrch StatusBar
   @ maxrow(),00 say "Press <Alt+W> to enter WildSrch"
   @ maxrow(),60 say "WildSrch: OFF"


   // Default coordinates
   nTop    := IF(VALTYPE( nTop )  != "N", 0       , nTop)
   nLeft   := IF(VALTYPE( nLeft)  != "N", 0       , nLeft)
   nBottom := IF(VALTYPE(nBottom) != "N", MAXROW()-1, nBottom)
   nRight  := IF(VALTYPE(nRight)  != "N", MAXCOL(), nRight)

   // Was array of fields passed?
   lColFld := VALTYPE(aColFld) == "A"

   // Macro-compile any udf passed
   lUdf := (VALTYPE(cUdf) == "C" .AND. LEN(cUdf) > 0)
   bUdf := IF(lUdf, &("|nMode, nElement| " + cUdf +;
                    "(nMode, nElement)}" ), ;
                    {|| .T.})

   // Macro-compile any while conditions passed, else
   // default to every record
   bWhileCond := IF(VALTYPE(cWhileCond) == "C" .AND.;
               LEN(cWhileCond) > 0,;
               &("{|| " + cWhileCond + " }") , ;
               {|| .T.})

   // Was array of titles passed?
   lColTit := VALTYPE(aColTit) == "A"

   // Create Browse object
   o := TBROWSEDB( nTop, nLeft, nBottom, nRight )

   // Color table
   o:colorSpec := "B+/W, W+/G, GR+/R"

   // Add custom "Skippers"
   // To handle While condition
   o:skipBlock     := {|nSkip| SkipDB(nSkip, bWhileCond)}
   o:goTopBlock    := {|| GoTopDB( bWhileCond )}
   o:goBottomBlock := {|| GoBottomDB( bWhileCond )}

   // Determine number of columns
   nNumOfFlds := IF(lColFld, LEN(aColFld), FCOUNT())

   // Add Columns
   FOR nCount := 1 TO nNumOfFlds
      // Make object
      column := TBColumnNew( IF(lColTit, aColTit[ nCount ],;
               IF(lColFld, aColFld[ nCount ], ;
               FIELDNAME( nCount ) ) ), ;
               FIELDWBLOCK(IF(lColFld, ;
               aColFld[ nCount ], ;
               FIELDNAME( nCount ) ), SELECT() ) )
      column:headSep := HEADSEP
      column:colSep  := COLSEP
      column:footSep := FOOTSEP

      // Memo?
      // Then do not show data
      // Show memo designator
      column:block := IF(VALTYPE(;
                       &(FIELDNAME(nCount ))) == "M", ;
                       {|| "Memo"}, column:block)

      // Add column to Browse object
      o:addColumn(column)

   NEXT

   // Main loop
   nMode := 0
   o:gotop()
   WHILE .T.
      // Determine wheather to continue, and if so, in
      // what manner should the screen be repainted
      IF nRetUdf == 0
         EXIT

      ELSEIF nRetUdf == 1
         o:refreshCurrent()

      ELSEIF nRetUdf == 2
         o:refreshAll()

      ENDIF
      nRetUdf := 1

      // Stabilize
      WHILE !o:stabilize()
      END

      IF o:stable
         // Retain coordinates
         nRow := ROW()
         nCol := COL()

         IF o:hitTop .OR. o:hitBottom
            TONE(500, 1)

            IF o:hitTop .AND. o:hitBottom
               // Empty
               nMode := 3

            ELSEIF o:hitTop
               nMode := 1

            ELSEIF o:hitBottom
               nMode := 2

            ENDIF

         ENDIF

         // If there is an Udf to process, do so
         IF lUdf
            nRetUdf := EVAL( bUdf, nMode, o:colPos )

         ELSEIF nMode == 4 .AND. !lUdf
            // If non-movement key is pressed and no Udf
            // to process, then, get out of here
            EXIT

         ENDIF

         // Reset mode, if WAS greater than zero
         IF nMode > 0
            nMode := 0

         ENDIF

         // Get Keys
         nKey := INKEY(0)

         // Process Keys
         IF nKey == K_UP
            o:up()

         ELSEIF nKey == K_DOWN
            o:down()

         ELSEIF nKey == K_LEFT
            o:left()

         ELSEIF nKey == K_RIGHT
            o:right()

         ELSEIF nKey == K_PGUP
            o:pageUp()

         ELSEIF nKey == K_PGDN
            o:pageDown()

         ELSEIF nKey == K_CTRL_PGUP
            o:goTop()

         ELSEIF nKey == K_CTRL_PGDN
            o:goBottom()

         // WildSrch
         ELSEIF nKey == K_ALT_W

            if !lWildSrch
               wScreen := savescreen(10,10,14,60)
               aResult:={}
               cSrchArg:=space(20)
               DISPBOX(10,10,14,60, "Ŀ ")
               @ 11,12 say "To test it enter 'bells'.."
               @ 12,12 say "Search for:"
               setcursor(SC_NORMAL)
               @ 12,25 get cSrchArg
               read
               setcursor(SC_NONE)
               cSrchArg:=upper(trim(cSrchArg))
               aResult := WildSrch("test.ntx",cSrchArg)
               if !empty(aResult)
                  aResult:=aSort(aResult)     // Sorting recno's
                  lWildSrch:=.t.
                  o:skipBlock     := {|nSkip| wSkipDB(nSkip, { | | ascan(aResult,recno()) !=0} ) }
                  o:goTopBlock    := {|| wGoTopDB( { | | ascan(aResult,recno()) !=0} )}
                  o:goBottomBlock := {|| wGoBottomDB( { | | ascan(aResult,recno()) !=0} )}
                  o:goTop()
                  o:RefreshAll()
                  @ maxrow(),70 say "ON "
               endif
            else
               lWildSrch:=.f.
               o:skipBlock     := {|nSkip| SkipDB(nSkip, bWhileCond)}
               o:goTopBlock    := {|| GoTopDB( bWhileCond )}
               o:goBottomBlock := {|| GoBottomDB( bWhileCond )}
               //o:GoTop()
               o:RefreshAll()
               @ maxrow(),70 say "OFF"
            endif
            restscreen(10,10,14,60,wScreen)
         ELSE
            nMode := 4

         ENDIF

      ENDIF

   END

   SETCURSOR(nCursor)
   SETCOLOR(cColr)
   RETURN (NIL)

STATIC FUNCTION GoTopDB( bWhileCond )
   // Skip all the way to top of condition
   WHILE EVAL( bWhileCond ) .AND. !BOF()
      DBSKIP(-1)
      IF !EVAL( bWhileCond ) .OR. BOF()
         DBSKIP(1)
         EXIT

      END

   END
   RETURN (NIL)

STATIC FUNCTION GoBottomDB( bWhileCond )

   WHILE EVAL( bWhileCond) .AND. !EOF()
       DBSKIP(1)
       IF !EVAL( bWhileCond) .OR. EOF()
          DBSKIP(-1)
          EXIT

       END
   END
   RETURN (NIL)

STATIC FUNCTION SkipDB( nSkipRequest, bWhileCOnd )
   LOCAL nActSkip := 0

   IF nSkipRequest == 0
      DBSKIP(0)

   ELSEIF nSkipRequest > 0 .AND. !EOF()
      // keep skipping until number of skips requested
      // is satisfied

      WHILE nActSkip < nSkipRequest
         DBSKIP(1)

         // Evaluate condition to each record
         IF !EVAL( bWhileCond ) .OR. EOF()
            DBSKIP(-1)
            EXIT

         END

         // Mantain counter
         nActSkip++

      END

   ELSEIF nSkipRequest < 0

      WHILE nActSkip > nSkipRequest
         DBSKIP(-1)
         // This check is important! BOF bangs do not
         // sit on phanton record. Rather the top record
         // is evaluated, and may meet the condition
         IF BOF()
            EXIT

         ENDIF

         // Evaluate condition to each record
         IF !EVAL( bWhileCond ) .OR. BOF()
            DBSKIP(1)
            EXIT

         ENDIF

         // Mantain counter
         nActSkip--

      END

   ENDIF
   // Return the number of ACTUAL skips processed
   RETURN (nActSkip)

// Functions for WildSrch --------------------------------------------------
// These functions (with prefix "w") were called instead of the original
// functions when WILDSRCH is active.

STATIC FUNCTION wGoTopDB( bWhileCond )
   DBGOTOP()            // to top of File - sorted by index
   WHILE !EVAL( bWhileCond)
       DBSKIP(1)
   END

   RETURN (NIL)


STATIC FUNCTION wGoBottomDB( bWhileCond )
   DBGOBOTTOM()
   WHILE !EVAL( bWhileCond)
       DBSKIP(-1)
   ENDDO

   RETURN (NIL)

STATIC FUNCTION wSkipDB( nSkipRequest, bWhileCond )
   LOCAL nActSkip := 0
   LOCAL nRecord  := recno()    // last record in <bWhileCond>

   IF nSkipRequest == 0
      DBSKIP(0)

   ELSEIF nSkipRequest > 0 .AND. !EOF()
      // keep skipping until number of skips requested
      // is satisfied

      WHILE nActSkip < nSkipRequest
         DBSKIP(1)

         // Evaluate condition to each record
         IF !EVAL( bWhileCond )
            if EOF()            // EOF
               goto nRecord     // go back to last record that matches
               EXIT             // the <bWhileCond>
            endif
            loop
         END

         // Mantain counter
         nActSkip++
         nRecord:= recno()      // update last record in <bWhileCond>

         // The problem on an active index-file is to check the
         // "out-of-order"-condition. If the recno of the actual record
         // is higher than the highest recno in <aResult>, the skipper runs
         // to eof(). So always point nRecord to the LAST record that
         // matches <bWhileCond>. When eof() go back to nRecord!
         // If anyone find a better way send me an e-mail.

      END

   ELSEIF nSkipRequest < 0
      WHILE nActSkip > nSkipRequest
         DBSKIP(-1)
         // Evaluate condition to each record
         IF !EVAL( bWhileCond )
            if BOF()            // BOF
               goto nRecord     // go forward to last record that matches
               EXIT             // the <bWhileCond>
            endif
            loop
         ENDIF

         // Mantain counter
         nActSkip--
         nRecord:=recno()       // update last record in <bWhileCond>

         // The problem on an active index-file is to check the
         // "out-of-order"-condition. If the recno of the actual record
         // is lower than the lowest recno in <aResult>, the skipper runs
         // to bof(). So always point nRecord to the LAST record that
         // matches <bWhileCond>. When bof() go back to nRecord!
         // If anyone find a better way send me an e-mail.

      END

   ENDIF
   // Return the number of ACTUAL skips processed
   RETURN (nActSkip)

// EOF - TBR20.PRG //
