/*  WILDSRCH.PRG 

   AUTHOR : Rainer Kppl-Schfer
   DATE   : Fri  3/13/98
   NOTICE : Copyright (c) 1997 - 1998 by
            RKS-Software
            Rainer Kppl-Schfer
            Goldbergstrasse 9
            D-71065 Sindelfingen

            All Rights Reserved
   VERSION: Clipper 5.2e
   PURPOSE: Fast Search in .NTX-Files
   Notes  : Released to public domain for use with attribution

   Version       : 1.1

   Compiling     :  CLIPPER wildsrch /a /n /w /l


   What happens  :  if you have a file with articles, customers or.. or..
   inside           or.. - and you have indexed it - have a nice
   WILSDRCH ?       search . But what happens, when you're looking for
                    something, that is not at the beginning of the index-
                    key, ie. KEY <Christmas Bells>, you search for
                    <bells>  - you will get a "key not found".
                    Try it with "set filter to.." - and a long time
                    later....

                    The WILDSRCH-Function is scanning directly the
                    NTX-File - much smaller than yor DBF-file.

                    I use the low-level-functions to open and read the
                    NTX-file.

                    Each NTX-File is organized in pages of 1K. The first
                    page is the header-page - here we need only three in-
                    formations:

                    - The Item-Size of each item
                    - The key-size of the index-key
                    - The maximum number of items on each index-page

                    Then we have a look at each index-page
                    At offset <h00> on each page we find the unsigned
                    item_count - how many items this particular page holds.
                    For each item will follow the item-offset - holding the
                    offsets within the page, where the items are located
                    (at h02, h04, h06 ... )

                    Each item has the following structure:

                    len: 8 + key-size

                    long page-offset - file offset (within NTX-file) where
                                       sub-tree is. if there's no sub-tree,
                                       a 0 is stored.
                    long recno       - record number for this key value
                                       (in the DBF-file)
                    char key-value   - the key-value, size is defined in
                                       header-page.


                    WILDSRCH first read the header-page, get the information,
                    and then read all following index-pages.
                    At each page first get the information at h00 (how
                    many items at this page), then read for each item
                    the file-offset (at h02, h04, h06 ...), then read
                    the key-value of this item. If the search-argument
                    matches the key-value, extract the record-number (inside
                    the DBF-file) and store it in the search-result-array,
                    then check the next item and so on..
                    If all items were passed, read the next page and
                    so on..

                    After reading all pages return the result-array.
                    Thats's all.

                    NOT included: - a ascending or descending read of the
                                  NTX-file - the search-results are not
                                  sorted

   Restrictions:    WILDSRCH only can read NTX-files - not a CDX- or other
                    Index-Files.
                    The DBF-file and the corresponding NTX-files must be
                    opened in SHARED-Mode (enables FOPEN to open the NTX-File)

   Speeeeeed   :    I use this function in a commercial application to
                    look for articles in a NTX-File. The key-value is 40
                    characters long, the database holds 12,517 articles.
                    To search for any expression the function needs 0.82 -
                    1.42 seconds (smartdrv works..) - a little bit faster
                    than the "set filter"-command.

   How to Use  :    In a TBROWSE you can create a SKIP-Block, that checks, if
                    the actual record is included in the result-array.
                    See in TBR20A.PRG

   NTX-Info    :    For more information about NTX-files read the file
                    NTX.TXT - a original work by Cesar A. Gil - i found it
                    at   WWW.WOTZIT.DEMON.CO.UK
                    Thanks, Cesar
   
*/

#include "fileio.ch"

FUNCTION WildSrch (cNTXFile,cSrchArg,nMaxHits) // ------------------------------
/*
parameters    :  cNTXFile  : drive:\path\filename.ext
                             (the NTX-File, to be scanned)
                 cSrchArg  : character-string with the search-argument

                 nMaxHits  : If the result-array is as long as <nMaxHits>
                             the function returns.
                             If nMaxHits = NIL, nMaxHits is set to 4,096
                             (Array is full..)

returns       :  aResult   : an array with the record-numbers who
                             matched <cSrchArg>
*/


// Search Result
LOCAL aResult   := {}               // array to hold the record-numbers

// For Low-Level-Services
LOCAL nBlock    :=1024              // Size of each index-page
LOCAL cBuffer   :=space(nBlock)     // Buffer for FREAD
LOCAL nHandle   :=-1                // file-handle
LOCAL nLen      :=0                 // length of index-file
LOCAL nPages    :=0                 // index-pages (excl. header)


// Index-Header (offset from 0, substr from 1 -> always (+1) )
LOCAL nPtrItemSize   := 13    // h0C -> Pointer to ItemSize (keysize+8)
LOCAL nPtrKeySize    := 15    // h0E -> Pointer to KeySize
LOCAL nPtrMaxItem    := 19    // h12 -> Pointer to max. items on each page

LOCAL nLenItem       :=0      // length of each item
LOCAL nLenKey        :=0      // length of each key
LOCAL nItemMax       :=0      // max. items on each page

// IndexPage   (offset from 0, substr form 1 -> always (+1) )
LOCAL nPtrItemPage   := 1     // h00 -> Pointer to Items on this page

LOCAL nItemCnt       := 0     // items on this page

LOCAL x:=0, y:=0              // FOR-loop indices
LOCAL nPtrItem       :=0      // pointer to current item
LOCAL nItemOffset    :=0      // offset of current item
LOCAL cKey           :=""     // key-value of current item
LOCAL nRecord        :=0      // recno of current item (inside the DBF-File)

// Check nMaxHits and Set Break
LOCAL lBreak    := .f.        // Break-Condition

nMaxHits := if(nMaxHits==NIL,4096,nMaxHits)

// Let's do..

// FileOpen

nHandle := fopen(cNTXFile, FO_READ + FO_SHARED)

// Calculate how many pages excl. header-page

nLen    := fseek(nHandle, 0, FS_END)
nPages  := int(nLen/nBlock) - 1
fseek(nHandle,0)

// Read Header and get some information

fread(nHandle, @cBuffer, nBlock )

nLenItem := BIN2W( substr(cBuffer,nPtrItemSize,2) )
nLenKey  := BIN2W( substr(cBuffer,nPtrKeySize,2) )
nItemMax := BIN2W( substr(cBuffer,nPtrMaxItem,2) )

//Now Read all Index-Pages and check for Search-Value

FOR X := 1 to nPages

    fread(nHandle, @cBuffer, nBlock)

    // check how many items on this page
    nItemCnt := BIN2W( substr(cBuffer, nPtrItemPage, 2) )

    // check all items on this page
    FOR Y := 1 TO nItemCnt

        // Page-Offset to read Item-Offset
        nPtrItem := (y*2)+1

        // Read Item-Offset
        nItemOffset := BIN2W( substr(cBuffer, nPtrItem, 2) ) + 1

        // Check current item
        cKey := substr(cBuffer, nItemOffset+8, nLenKey)

        if at(cSrchArg, cKey) != 0   // SearchArgument found
           // Read recno from index-page (beginning at <nItemOffset+4>)
           nRecord  := BIN2L( substr(cBuffer, nItemOffset+4, 4) )

           // Add recno to result-array
           aadd(aResult, nRecord)

           // Check nMaxHits
           if len(aResult) >= nMaxHits
              lBreak := .t.
              EXIT
           endif

        endif

    NEXT Y

    if lBreak
       EXIT
    endif

NEXT X

// close NTX-File
fclose(nHandle)

RETURN(aResult)   // EOF WildSrch ----------------------------------------------
